from __future__ import annotations

import logging
import typing

import numpy as np
from PySide2 import QtWidgets, QtGui, QtCore

from qt_extensions.combobox import QComboBox
from qt_extensions.icons import MaterialIcon
from qt_extensions.parameters import FloatParameter


logger = logging.getLogger(__name__)
CHANNELS = ['rgba', 'red', 'green', 'blue', 'alpha']


class GraphicsItem(QtWidgets.QGraphicsItem):
    def __init__(self, parent: QtWidgets.QGraphicsItem | None = None) -> None:
        super().__init__(parent)
        self.image = QtGui.QImage()

    def paint(
        self,
        painter: QtGui.QPainter,
        option: QtWidgets.QStyleOptionGraphicsItem,
        widget: QtWidgets.QWidget | None = None,
    ) -> None:
        painter.drawImage(option.rect, self.image)

    def boundingRect(self) -> QtCore.QRectF:
        rect = QtCore.QRectF(self.image.rect())
        return rect


class GraphicsScene(QtWidgets.QGraphicsScene):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._item: GraphicsItem | None = None

        # bounding box frame
        rect = QtCore.QRect()
        color = self.palette().color(QtGui.QPalette.Midlight)
        pen = QtGui.QPen(color)
        pen.setStyle(QtCore.Qt.DotLine)
        pen.setWidth(0)
        brush = QtGui.QBrush()
        brush.setStyle(QtCore.Qt.NoBrush)

        self._frame = self.addRect(rect, pen, brush)
        self._frame.setZValue(1)

        self.set_background_color(QtGui.QColor(0, 0, 0))

    def item(self) -> GraphicsItem:
        return self._item

    def set_background_color(self, value: QtGui.QColor) -> None:
        self.setBackgroundBrush(QtGui.QBrush(value))

    def set_item(self, value: GraphicsItem) -> None:
        if self._item and self._item.parent() == self:
            self.removeItem(self._item)
        self._item = value
        self.addItem(value)

    def update_frame(self, resolution: QtCore.QSize) -> None:
        rect = QtCore.QRect(QtCore.QPoint(), resolution)
        self._frame.setRect(rect)


class GraphicsView(QtWidgets.QGraphicsView):
    zoom_changed = QtCore.Signal(float)
    position_changed = QtCore.Signal(QtCore.QPoint)
    pixel_position_changed = QtCore.Signal(QtCore.QPoint)
    pixel_color_changed = QtCore.Signal(QtGui.QColor)

    # initialize a 16k scene rect
    scene_rect = QtCore.QRect(-(2**13), -(2**13), 2**14, 2**14)

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._dragging: bool = False

        self.setMouseTracking(True)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setDragMode(self.ScrollHandDrag)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setSceneRect(self.scene_rect)
        self.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.viewport().setCursor(QtCore.Qt.CrossCursor)

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        if event.key() == QtCore.Qt.Key_F:
            self.fit()
            event.accept()
            return
        super().keyPressEvent(event)

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        if event.button() == QtCore.Qt.MidButton:
            handmade_event = QtGui.QMouseEvent(
                QtCore.QEvent.MouseButtonPress,
                QtCore.QPointF(event.pos()),
                QtCore.Qt.LeftButton,
                event.buttons(),
                QtCore.Qt.KeyboardModifiers(),
            )
            super().mousePressEvent(handmade_event)
            self.viewport().setCursor(QtCore.Qt.CrossCursor)

        if event.button() == QtCore.Qt.LeftButton:
            self._dragging = True
            self.mouseMoveEvent(event)
            event.accept()
            return

        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        if event.button() == QtCore.Qt.MidButton:
            release_event = QtGui.QMouseEvent(
                QtCore.QEvent.MouseButtonRelease,
                QtCore.QPointF(event.pos()),
                QtCore.Qt.LeftButton,
                event.buttons(),
                QtCore.Qt.KeyboardModifiers(),
            )
            super().mouseReleaseEvent(release_event)
            self.viewport().setCursor(QtCore.Qt.CrossCursor)

        if event.button() == QtCore.Qt.LeftButton:
            self._dragging = False
            event.accept()
            return

        super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        super().mouseMoveEvent(event)

        if self.scene() and self.scene().item():
            cursor_position = self.mapToScene(event.pos())
            bounding_rect = self.scene().item().boundingRect()
            position = QtCore.QPoint(
                np.floor(cursor_position.x()),
                np.floor(bounding_rect.height() - cursor_position.y()),
            )
            if self._dragging:
                self.position_changed.emit(position)
            self.pixel_position_changed.emit(position)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        zoom_in_factor = 1.25
        zoom_out_factor = 1 / zoom_in_factor

        # zoom
        if event.angleDelta().y() > 0:
            zoom_factor = zoom_in_factor
        else:
            zoom_factor = zoom_out_factor
        self.scale(zoom_factor, zoom_factor)

        self.zoom_changed.emit(self.absolute_scale())
        event.accept()

    def absolute_scale(self) -> float:
        # NOTE: since there will never be rotation, and scale in x and y are the same,
        # m11 can be used as scale
        return self.transform().m11()

    def fit(self) -> None:
        if self.scene() and self.scene().item():
            self.fitInView(self.scene().item(), QtCore.Qt.KeepAspectRatio)
            self.zoom_changed.emit(self.absolute_scale())

    def set_absolute_scale(self, value: float) -> None:
        self.setTransform(QtGui.QTransform.fromScale(value, value))

    def zoom(self, factor: float) -> None:
        if factor == 0:
            self.fit()
        else:
            self.set_absolute_scale(factor)


class Footer(QtWidgets.QWidget):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._init_ui()

        self.set_background_color(QtGui.QColor(0, 0, 0))

    def _init_ui(self) -> None:
        self.setLayout(QtWidgets.QHBoxLayout())

        self.setAutoFillBackground(True)

        self.resolution_lbl = QtWidgets.QLabel('resolution')
        self.layout().addWidget(self.resolution_lbl)

        self.layout().addStretch()

        self.coordinates_lbl = QtWidgets.QLabel('coordinates')
        self.layout().addWidget(self.coordinates_lbl)

        self.rgb_lbl = QtWidgets.QLabel('rgb')
        self.layout().addWidget(self.rgb_lbl)

        self.hsv_lbl = QtWidgets.QLabel('hsv')
        self.layout().addWidget(self.hsv_lbl)

    def set_background_color(self, value: QtGui.QColor) -> None:
        palette = self.palette()
        palette.setColor(QtGui.QPalette.Window, value)
        palette.setColor(QtGui.QPalette.WindowText, QtGui.QColor(255, 255, 255))
        self.setPalette(palette)

    def update_pixel_color(self, color: QtGui.QColor | None) -> None:
        if color.isValid():
            r, g, b, a = color.getRgbF()
            rgb = (
                f'<font color="#ff2222">{r:.4f}</font> '
                f'<font color="#00ff22">{g:.4f}</font> '
                f'<font color="#0088ff">{b:.4f}</font>'
            )
            h, s, v, a = color.getHsvF()
            h = max(h, 0)
        else:
            rgb = ''
            h, s, v = 0, 0, 0
        hsv = f'H: {h:.2f} S: {s:.2f} V: {v:.2f}'

        self.rgb_lbl.setText(rgb)
        self.hsv_lbl.setText(hsv)

    def update_pixel_position(self, position: QtCore.QPoint | None) -> None:
        if position is not None:
            coordinates = f'x={position.x()} y={position.y()}'
        else:
            coordinates = ''
        self.coordinates_lbl.setText(coordinates)

    def update_resolution(self, resolution: QtCore.QSize) -> None:
        text = f'{resolution.width():.0f}x{resolution.height():.0f}'
        self.resolution_lbl.setText(text)


class ToolBar(QtWidgets.QToolBar):
    refreshed: QtCore.Signal = QtCore.Signal()
    paused: QtCore.Signal = QtCore.Signal(bool)
    exposure_changed: QtCore.Signal = QtCore.Signal(float)
    zoom_changed: QtCore.Signal = QtCore.Signal(float)

    pause_color = QtGui.QColor(217, 33, 33)

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._zoom = 0
        self._exposure = 0
        self._exposure_cache = 0

        self._init_actions()

    def _init_actions(self) -> None:
        size = self.style().pixelMetric(QtWidgets.QStyle.PM_SmallIconSize)
        self.setIconSize(QtCore.QSize(size, size))

        self.channel_cmb = QComboBox()
        self.channel_cmb.keyPressEvent = lambda event: event.ignore()
        self.channel_cmb.addItems(CHANNELS)
        channel_action = QtWidgets.QWidgetAction(self)
        channel_action.setText('channel')
        channel_action.setDefaultWidget(self.channel_cmb)
        self.addAction(channel_action)

        # exposure toggle
        icon = MaterialIcon('toggle_on')
        icon_off = MaterialIcon('toggle_off')
        palette = self.palette()
        color = palette.color(QtGui.QPalette.Highlight)
        pixmap = icon_off.pixmap(0, QtGui.QIcon.Active, QtGui.QIcon.On, color)
        icon.addPixmap(pixmap, QtGui.QIcon.Active, QtGui.QIcon.On)

        self.exposure_toggle_action = QtWidgets.QAction(icon, 'exposure_toggle', self)
        self.exposure_toggle_action.setCheckable(True)
        self.exposure_toggle_action.toggled.connect(self._exposure_toggled)
        self.addAction(self.exposure_toggle_action)

        # exposure slider
        self.exposure_slider = FloatParameter(parent=self)
        self.exposure_slider.set_slider_min(-10)
        self.exposure_slider.set_slider_max(10)
        self.exposure_slider.value_changed.connect(self._exposure_changed)
        palette = self.exposure_slider.slider.palette()
        palette.setColor(QtGui.QPalette.Highlight, palette.color(QtGui.QPalette.Base))
        self.exposure_slider.slider.setPalette(palette)

        exposure_action = QtWidgets.QWidgetAction(self)
        exposure_action.setText('exposure')
        exposure_action.setDefaultWidget(self.exposure_slider)
        self.addAction(exposure_action)

        # refresh
        icon = MaterialIcon('refresh')
        refresh_action = QtWidgets.QAction(icon, 'refresh', self)
        refresh_action.triggered.connect(self.refreshed.emit)
        self.addAction(refresh_action)

        # pause
        icon = MaterialIcon('pause')
        color = self.pause_color
        pixmap = icon.pixmap(0, QtGui.QIcon.Active, QtGui.QIcon.On, color)
        icon.addPixmap(pixmap, QtGui.QIcon.Active, QtGui.QIcon.On)
        pause_action = QtWidgets.QAction(icon, 'pause', self)
        pause_action.setCheckable(True)
        pause_action.toggled.connect(self.paused.emit)
        self.addAction(pause_action)

        # zoom
        self.zoom_cmb = QComboBox()
        self.zoom_cmb.addItem('fit')
        factors = [0.10, 0.25, 0.33, 0.5, 0.75, 1, 1.5, 2, 3, 4, 5]
        for factor in reversed(factors):
            self.zoom_cmb.addItem(f'{factor:2.0%}', factor)
        self.zoom_cmb.setMaxVisibleItems(self.zoom_cmb.count())
        self.zoom_cmb.currentIndexChanged.connect(self._zoom_index_changed)

        # NOTE: currently AdjustToContents doesn't do anything, but this might be because
        # the placeholder text is broken, thus setting setMinimumContentsLength works for
        # ensuring that the full placeholder text is visible.
        self.zoom_cmb.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.zoom_cmb.setMinimumContentsLength(6)

        zoom_action = QtWidgets.QWidgetAction(self)
        zoom_action.setText('zoom')
        zoom_action.setDefaultWidget(self.zoom_cmb)
        self.addAction(zoom_action)

    def exposure(self) -> float:
        return self._exposure

    def find_action(self, text: str) -> QtWidgets.QAction | None:
        for action in self.actions():
            if action.text() == text:
                return action

    def set_channel(self, channel: str) -> None:
        self.channel_cmb.setCurrentText(channel)

    def set_exposure(self, exposure: float) -> None:
        self._exposure = exposure
        self.exposure_slider.set_value(exposure)

    def set_zoom(self, zoom: float) -> None:
        self._zoom = zoom
        self.zoom_cmb.setCurrentIndex(-1)
        self.zoom_cmb.setPlaceholderText(f'{zoom:2.1%}')

    def zoom(self) -> float:
        return self._zoom

    def _exposure_changed(self, value: float) -> None:
        self.exposure_toggle_action.blockSignals(True)
        self.exposure_toggle_action.setChecked(value != 0)
        self.exposure_toggle_action.blockSignals(False)

        self._exposure = value
        if value != 0:
            self._exposure_cache = value
        self.exposure_changed.emit(value)

    def _exposure_toggled(self) -> None:
        exposure = self._exposure_cache if self.exposure() == 0 else 0
        self.set_exposure(exposure)

    def _zoom_index_changed(self, index: int) -> None:
        if self.zoom_cmb.currentText() == 'fit':
            self._zoom = 0
        elif index > 0:
            self._zoom = self.zoom_cmb.currentData()
        else:
            return
        self.zoom_changed.emit(self._zoom)


class Viewer(QtWidgets.QWidget):
    refreshed: QtCore.Signal = QtCore.Signal()
    pause_changed: QtCore.Signal = QtCore.Signal(bool)
    position_changed: QtCore.Signal = QtCore.Signal(QtCore.QPoint)
    channel_changed: QtCore.Signal = QtCore.Signal(str)

    background_color = QtGui.QColor(0, 0, 0)
    pause_color = QtGui.QColor(217, 33, 33)

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self.paused = False
        self._resolution = QtCore.QSize()
        self._channel: str = CHANNELS[0]
        self._exposure: float = 0
        self._array = np.ndarray((0, 0, 3), np.float32)

        self.post_processes: list[typing.Callable] = [
            self._expose_image,
            self._select_channel,
        ]

        self._init_ui()

    def _init_ui(self) -> None:
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        # toolbar
        self.toolbar = ToolBar()
        self.toolbar.pause_color = self.pause_color
        self.toolbar.refreshed.connect(self.refresh)
        self.toolbar.paused.connect(self.pause)
        self.toolbar.exposure_changed.connect(self._exposure_changed)
        self.layout().addWidget(self.toolbar)

        # view
        self.item = GraphicsItem()

        self.scene = GraphicsScene()
        self.scene.set_background_color(self.background_color)
        self.scene.set_item(self.item)

        self.view = GraphicsView()
        self.view.setScene(self.scene)
        self.view.fit()
        self.view.zoom_changed.connect(self.toolbar.set_zoom)
        self.toolbar.zoom_changed.connect(self.view.zoom)
        self.layout().addWidget(self.view)

        # footer
        self.footer = Footer()
        self.footer.set_background_color(self.background_color)
        self.layout().addWidget(self.footer)

        # signals
        self.view.pixel_position_changed.connect(self._pixel_position_changed)
        self.view.position_changed.connect(self.position_changed.emit)

    def keyPressEvent(self, event: QtCore.QEvent) -> None:
        channels = {
            QtCore.Qt.Key_R: CHANNELS[1],
            QtCore.Qt.Key_G: CHANNELS[2],
            QtCore.Qt.Key_B: CHANNELS[3],
            QtCore.Qt.Key_A: CHANNELS[4],
        }
        channel = channels.get(event.key())
        if channel:
            if channel == self._channel:
                channel = 'rgba'
            self.set_channel(channel)
        else:
            super().keyPressEvent(event)
            return
        event.accept()

    def color_at(self, position: QtCore.QPoint) -> QtGui.QColor:
        height, width = self._array.shape[:2]
        x = position.x()
        y = height - 1 - position.y()
        if x < 0 or x >= width or y < 0 or y >= height:
            color = QtGui.QColor()
            color.convertTo(QtGui.QColor.Invalid)
        else:
            rgb = self._array[y, x]
            color = QtGui.QColor.fromRgbF(*rgb)
        return color

    def channel(self) -> str:
        return self._channel

    def exposure(self) -> float:
        return self._exposure

    def pause(self, state=True) -> None:
        self.paused = state

        if self.paused:
            self.view.setFrameShape(QtWidgets.QFrame.Box)
            self.view.setStyleSheet(
                f'QFrame {{ border: 1px solid {self.pause_color.name()}; }}'
            )
            self.view.setEnabled(False)
        else:
            self.view.setFrameShape(QtWidgets.QFrame.NoFrame)
            self.view.setStyleSheet('')
            self.view.setEnabled(True)

        self.pause_changed.emit(self.paused)

    def refresh(self) -> None:
        self.refreshed.emit()

    def relative_position(self, position: QtCore.QPoint) -> QtCore.QPointF:
        resolution = self.resolution()
        return QtCore.QPointF(
            (position.x() / resolution.width() - 0.5) * 2,
            (position.y() / resolution.height() - 0.5) * 2,
        )

    def resolution(self) -> QtCore.QSize:
        return self._resolution

    def set_array(self, array: np.ndarray) -> None:
        array = self._array_as_image(array)

        if self.paused:
            return

        self._array = array
        height, width = array.shape[:2]

        self._update_image()

        # trigger fit to view
        self.set_resolution(QtCore.QSize(width, height))

    def set_channel(self, channel: str) -> None:
        self._channel = channel
        self.toolbar.set_channel(channel)
        self._channel_changed(channel)

    def set_exposure(self, exposure: float) -> None:
        self._exposure = exposure
        self.toolbar.set_exposure(exposure)
        self._exposure_changed(exposure)

    def set_resolution(self, resolution: QtCore.QSize) -> None:
        if self._resolution != resolution:
            self._resolution = resolution
            self.footer.update_resolution(resolution)
            self.scene.update_frame(resolution)
            self.view.fit()

    def set_state(self, state: dict) -> None:
        values = {'exposure': 0}
        values.update(state)
        self.set_exposure(values['exposure'])

    def state(self) -> dict:
        state = {'exposure': self.exposure()}
        return state

    # noinspection PyMethodMayBeStatic
    def _array_as_image(self, array: np.ndarray) -> np.ndarray:
        # checks whether the array has either 1, 3 or 4 channels and converts
        # it to a 3 channel array while this is the only supported format

        if len(array.shape) == 2:
            array = np.dstack((array, array, array))
            return array
        if len(array.shape) == 3:
            if array.shape[2] > 3:
                array = array[:, :, :3]
                return array
            elif array.shape[2] == 3:
                return array
            elif array.shape[2] == 1:
                array = np.dstack((array[:, :, 0], array[:, :, 0], array[:, :, 0]))
                return array
        raise ValueError('Expected numpy array with either 1, 3 or 4 channels.')

    def _channel_changed(self, channel: str) -> None:
        self._channel = channel
        self._update_image()

    def _exposure_changed(self, value: float) -> None:
        if not self.paused:
            self._exposure = value
            self._update_image()

    def _expose_image(self, array: np.ndarray) -> None:
        gain = pow(2, self.exposure())
        np.multiply(array, gain, out=array)

    def _pixel_position_changed(self, position: QtCore.QPoint) -> None:
        self.footer.update_pixel_position(position)
        color = self.color_at(position)
        self.footer.update_pixel_color(color)

    def _select_channel(self, array: np.ndarray) -> None:
        index = CHANNELS.index(self.channel()) - 1
        if index == -1:
            return
        if index >= array.shape[2]:
            logger.debug('Image does not have that channel.')
            return
        for c in range(array.shape[2]):
            array[:, :, c] = array[:, :, index]

    def _update_image(self):
        height, width, channels = self._array.shape
        if not height or not width:
            return

        array = self._array.copy()

        for post_process in self.post_processes:
            post_process(array)

        np.clip(array, 0, 1, out=array)
        np.multiply(array, 255, out=array)
        array = array.astype(np.uint8)

        bytes_per_line = width * channels * array.dtype.itemsize
        image = QtGui.QImage(
            array.data, width, height, bytes_per_line, QtGui.QImage.Format_RGB888
        )
        # QImage is only valid as long as array stays in memory, so it is copied
        self.item.image = image.copy()
        self.item.update()



from __future__ import annotations

import dataclasses
import typing
from collections import OrderedDict
from functools import partial
from typing import Union

from PySide2 import QtCore, QtGui, QtWidgets

from qt_extensions import helper
from qt_extensions.icons import MaterialIcon
from qt_extensions.typeutils import cast, basic


@dataclasses.dataclass()
class DockWidgetState:
    current_index: int
    widgets: list[tuple[str, str]]
    detachable: bool
    auto_delete: bool
    is_center_widget: bool
    geometry: QtCore.QRect = QtCore.QRect()
    flags: int = 0


@dataclasses.dataclass()
class SplitterState:
    sizes: list[int]
    orientation: QtCore.Qt.Orientation
    states: list[typing.Union[DockWidgetState, 'SplitterState', None]]
    geometry: QtCore.QRect = QtCore.QRect()
    flags: int = 0


@dataclasses.dataclass()
class RegisteredWidget:
    cls: type
    title: str
    unique: bool


class DockTabBar(QtWidgets.QTabBar):
    detach_started: QtCore.Signal = QtCore.Signal(int)
    detach_moved: QtCore.Signal = QtCore.Signal(QtCore.QPoint)
    detach_finished: QtCore.Signal = QtCore.Signal(QtCore.QPoint)

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._drag_index = None
        self._detaching = False

        self.tabBarClicked.connect(self._tab_bar_click)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        if self._detaching:
            # mouse is pressed down and a tab is detached
            self.detach_moved.emit(event.pos())
        elif not self.rect().contains(event.pos()) and self._drag_index is not None:
            # a tab is about to be detached
            self.detach_started.emit(self._drag_index)
            self._detaching = True
        else:
            # no tab is detached
            # this must only be called when _detaching == False
            # undocking tabs while mouse move events are being processed leads to crashes
            # because of the tab QPainter events
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        if self._detaching:
            self.detach_finished.emit(event.pos())
        self._detaching = False
        self._drag_index = None

    def tabInserted(self, index: int) -> None:
        self._update_tab(index)

    def _request_tab_close(self, action: QtWidgets.QAction) -> None:
        for index in range(self.count()):
            button = self.tabButton(index, QtWidgets.QTabBar.RightSide)
            if action == button.defaultAction():
                self.tabCloseRequested.emit(index)
                return

    def _tab_bar_click(self, index: int) -> None:
        self._drag_index = index

    def _update_tab(self, index: int) -> None:
        close_icon = MaterialIcon('close')
        size = self.style().pixelMetric(QtWidgets.QStyle.PM_SmallIconSize)
        icon_size = QtCore.QSize(size, size)

        close_action = QtWidgets.QAction()
        close_action.setIcon(close_icon)
        close_action.triggered.connect(partial(self._request_tab_close, close_action))
        close_button = QtWidgets.QToolButton(self)
        close_button.setAutoRaise(True)
        close_button.setDefaultAction(close_action)
        close_button.setMaximumSize(icon_size)
        self.setTabButton(index, QtWidgets.QTabBar.RightSide, close_button)


class Splitter(QtWidgets.QSplitter):
    def childEvent(self, event: QtCore.QChildEvent) -> None:
        super().childEvent(event)
        if event.removed() and not self.count():
            self.deleteLater()


class DockWidget(QtWidgets.QTabWidget):
    dock_areas = (
        QtCore.Qt.LeftDockWidgetArea,
        QtCore.Qt.RightDockWidgetArea,
        QtCore.Qt.TopDockWidgetArea,
        QtCore.Qt.BottomDockWidgetArea,
        QtCore.Qt.NoDockWidgetArea,
    )

    def __init__(self, dock_window, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent or dock_window)

        self.dock_window = dock_window
        self.detachable = True
        self.auto_delete = True

        self._drag_widget = None
        self._hidden = False

        self._init_ui()

    def _init_ui(self) -> None:
        self.setTabBar(DockTabBar())
        self.setMovable(True)
        self.setTabsClosable(True)

        self.tabBar().detach_started.connect(self._detach_start)
        self.tabBar().detach_moved.connect(self._detach_move)
        self.tabBar().detach_finished.connect(self._detach_finish)

        self.tabCloseRequested.connect(self.close_tab)
        self.currentChanged.connect(self.update_window_title)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        # since widgets are stored in window._widgets, trigger garbage collection
        for widget in self.widgets().values():
            widget.deleteLater()
        super().closeEvent(event)

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        if self.isWindow():
            self.center()
        super().showEvent(event)

    def tabRemoved(self, index: int) -> None:
        self.try_delete()

    def add_dock_widget(
        self, widget: QtWidgets.QTabWidget, area: QtCore.Qt.DockWidgetArea
    ) -> None:
        if area == QtCore.Qt.NoDockWidgetArea:
            self.addTab(widget.widget(0), widget.tabText(0))
        else:
            parent = self.parent()

            if not parent:
                return

            if self.isWindow() or not isinstance(parent, QtWidgets.QSplitter):
                splitter = Splitter(QtCore.Qt.Vertical)

                if self.isWindow():
                    splitter.setParent(parent)
                    splitter.setWindowFlags(self.windowFlags())
                    splitter.show()
                elif isinstance(parent, QtWidgets.QScrollArea):
                    parent.setWidget(splitter)
                elif isinstance(parent.layout(), QtWidgets.QLayout):
                    parent.layout().replaceWidget(self, splitter)
                    splitter.setParent(parent)
                else:
                    # for other parents it's unknown on how to replace a widget
                    splitter.deleteLater()
                    return

                splitter.setGeometry(self.geometry())
                splitter.addWidget(self)
                parent = splitter

            orientation = area_orientation(area)
            index = parent.indexOf(self)
            if orientation != parent.orientation():
                if parent.count() == 1:
                    parent.setOrientation(orientation)
                else:
                    splitter = Splitter(orientation)
                    parent.insertWidget(index, splitter)
                    splitter.setParent(parent)
                    splitter.addWidget(self)
                    parent = splitter
                    index = 0

            if area in (
                QtCore.Qt.LeftDockWidgetArea,
                QtCore.Qt.TopDockWidgetArea,
            ):
                parent.insertWidget(index, widget)
            elif area in (
                QtCore.Qt.RightDockWidgetArea,
                QtCore.Qt.BottomDockWidgetArea,
            ):
                parent.insertWidget(index + 1, widget)

    def center(self) -> None:
        if not self.dock_window:
            return

        center = self.dock_window.geometry().center()
        x = center.x() - (self.width() // 2)
        y = center.y() - (self.height() // 2)
        self.move(x, y)

    def close_tab(self, index: int) -> None:
        widget = self.widget(index)
        self.removeTab(index)
        # since widgets are stored in window._widgets, trigger garbage collection
        if widget is not None:
            widget.deleteLater()

    def detach(self, index: int, interactive=False) -> None:
        if index not in range(self.count()) or not self.detachable:
            return

        geometry = self.geometry()

        if not self.isWindow():
            top_left = self.parent().mapToGlobal(geometry.topLeft())
            geometry.moveTopLeft(top_left)

        title = self.tabText(index)
        widget = self.widget(index)

        self._drag_widget = self.__class__(self.dock_window)
        self._drag_widget.setParent(self.dock_window)
        # setting window flag after parent creates a window
        self._drag_widget.float()
        # adding tab after setting window flag, triggers window title update
        self._drag_widget.addTab(widget, title)
        self._drag_widget.setGeometry(geometry)
        if interactive:
            self._drag_widget.setWindowOpacity(0.5)
        self._drag_widget.raise_()
        self._drag_widget.show()
        self._drag_widget.activateWindow()

    def dock_rects(self) -> dict[QtCore.Qt.DockWidgetArea, QtCore.QRect]:
        if not self._hidden:
            rects = {area: self._dock_rect(area) for area in self.dock_areas}
        else:
            rects = {}
        return rects

    def dock_preview_rect(self, area: QtCore.Qt.DockWidgetArea) -> QtCore.QRect:
        return self._dock_rect(area, 0.5)

    def float(self) -> None:
        self.setWindowFlag(QtCore.Qt.Tool, True)

    def update_window_title(self, index: int) -> None:
        if self.window() != self.dock_window:
            self.window().setWindowTitle(self.tabText(index))

    def try_delete(self) -> None:
        if self.auto_delete and not self.count():
            # this widget must not be deleted during a drag event
            # to get around this the widget or its parent will have their opacity set
            # to 0 or be hidden
            self._hidden = True
            if self._drag_widget:
                self._hide_recursively(self)
            else:
                self.deleteLater()

    def widgets(self) -> OrderedDict[str, QtWidgets.QWidget]:
        widgets = OrderedDict()
        for i in range(self.count()):
            widgets[self.tabText(i)] = self.widget(i)
        return widgets

    def _detach_start(self, index: int) -> None:
        self.detach(index, interactive=True)

    def _detach_move(self, position: QtCore.QPoint) -> None:
        if self._drag_widget:
            position = QtGui.QCursor().pos()
            height = self.style().pixelMetric(QtWidgets.QStyle.PM_TitleBarHeight)
            offset = position - QtCore.QPoint(height / 2, height / 2)
            self._drag_widget.move(offset)
            self.dock_window.add_dock_widget(self._drag_widget, position, True)

    def _detach_finish(self, position: QtCore.QPoint) -> None:
        if self._drag_widget:
            self._drag_widget.setWindowOpacity(1)
            position = QtGui.QCursor().pos()
            self.dock_window.add_dock_widget(self._drag_widget, position)
        self._drag_widget = None
        self.try_delete()

    def _dock_rect(
        self, area: QtCore.Qt.DockWidgetArea, factor: float = 0.2
    ) -> QtCore.QRect:
        size = self.size() * factor
        rect = self.rect()
        if area == QtCore.Qt.LeftDockWidgetArea:
            rect.setWidth(size.width())
            return rect
        elif area == QtCore.Qt.RightDockWidgetArea:
            right = rect.right()
            rect.setWidth(size.width())
            rect.moveRight(right)
            return rect
        elif area == QtCore.Qt.TopDockWidgetArea:
            rect.setHeight(size.height())
            return rect
        elif area == QtCore.Qt.BottomDockWidgetArea:
            bottom = rect.bottom()
            rect.setHeight(size.height())
            rect.moveBottom(bottom)
            return rect
        elif area == QtCore.Qt.NoDockWidgetArea:
            return rect
        return rect

    def _hide_recursively(self, widget) -> None:
        # hides the top most widget without deleting it
        if widget.isWindow():
            widget.setWindowOpacity(0)
        else:
            parent = widget.parent()
            if isinstance(parent, Splitter) and parent.count() == 1:
                # if splitter's only child is self, make it invisible
                self._hide_recursively(parent)
            elif parent:
                # if there are other children,
                # it's safe to hide as splitter will not auto delete
                widget.hide()


class DockWindow(QtWidgets.QWidget):
    widget_added: QtCore.Signal = QtCore.Signal(QtWidgets.QWidget)
    dock_widget_added: QtCore.Signal = QtCore.Signal(DockWidget)

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self.center_widget: DockWidget | None = None
        self.registered_widgets: dict[str, RegisteredWidget] = {}

        self._widgets: dict[str, QtWidgets.QWidget] = {}
        self._rubber_band: QtWidgets.QRubberBand | None = None

        self._init_rubber_band()
        self._init_ui()

    def _init_rubber_band(self) -> None:
        self._rubber_band = QtWidgets.QRubberBand(QtWidgets.QRubberBand.Rectangle)
        self._rubber_band.setParent(self)
        self._rubber_band.destroyed.connect(self._init_rubber_band)

    def _init_ui(self) -> None:
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        # center widget
        self.center_splitter = Splitter(QtCore.Qt.Vertical)
        self.center_widget = self._add_protected_dock_widget(self.center_splitter)
        self.layout().addWidget(self.center_splitter)

        self.layout().setStretch(0, 1)

    def add_dock_widget(
        self, widget: DockWidget, position: QtCore.QPoint, simulate: bool = False
    ) -> None:
        # add a widget at global position
        dock_rects = self._dock_rects()

        for target, rects in dock_rects.items():
            if target == widget:
                continue
            target_position = target.mapFromGlobal(position)
            if not target.rect().contains(target_position):
                self._rubber_band.hide()
                continue

            for area, rect in rects.items():
                if rect.contains(target_position):
                    if simulate:
                        preview_rect = target.dock_preview_rect(area)
                        self._rubber_band.setParent(target)
                        self._rubber_band.setGeometry(preview_rect)
                        self._rubber_band.show()
                    else:
                        self._rubber_band.hide()
                        target.add_dock_widget(widget, area)
                    return
                else:
                    self._rubber_band.hide()

    def create_dock_widget(
        self,
        widget: QtWidgets.QWidget | type | str,
        title: str | None = None,
    ) -> DockWidget:
        # helper function to turn QWidgets into DockWidgets
        for title_, widget_ in self._widgets.items():
            if widget == widget_:
                unique_title, widget_instance = title_, widget_
                break
        else:
            unique_title, widget_instance = self._add_widget(widget, title)

        dock_widget = DockWidget(dock_window=self)
        dock_widget.addTab(widget_instance, unique_title)
        self.dock_widget_added.emit(dock_widget)

        return dock_widget

    def dock_widgets(self) -> list[DockWidget]:
        # return all DockWidgets under this window. They are sorted in a manner
        # that makes sense for checking dock areas.
        children = self.findChildren(DockWidget)
        # floating widgets have the Tool window flag and are always in front:
        children = sorted(children, key=DockWidget.isWindow)
        # list deepest nested children first
        children = reversed(children)

        return list(children)

    def register_widget(
        self, cls: type, title: str | None = None, unique: bool = True
    ) -> RegisteredWidget:
        if title is None:
            title = helper.title(cls.__name__)

        key = cls.__name__
        if key not in self.registered_widgets:
            registered_widget = RegisteredWidget(cls, title, unique)
            self.registered_widgets[key] = registered_widget
        return self.registered_widgets[key]

    def unregister_widget(self, cls: type | str) -> None:
        # returns true if widget was found and unregistered, false if not found
        key = cls.__name__ if not isinstance(cls, type) else cls
        self.registered_widgets.pop(key, None)

    def set_window_state(self, state: dict) -> None:
        values = {'geometry': None, 'widgets': []}
        values.update(state)

        self.set_widget_states(values['widgets'])

        try:
            geometry = cast(QtCore.QRect, values['geometry'])
            self.setGeometry(geometry)
        except TypeError:
            pass

    def set_widget_states(
        self,
        states: list[dict],
    ) -> None:
        # store all current widgets for layout
        widgets = dict(self._widgets)
        # unparent all widgets to clean up layout
        for widget in widgets.values():
            widget.setParent(None)
            widget.close()

        states = cast(list[Union[DockWidgetState, SplitterState]], states)
        self._set_widget_states(states, self, widgets)

        # remove unused widgets
        for title, widget in widgets.items():
            widget.deleteLater()

    def state(self) -> dict:
        state = {'geometry': self.geometry(), 'widgets': self.widget_states()}
        state = basic(state)
        return state

    def widget_title(self, widget: QtWidgets.QWidget) -> str | None:
        for k, v in self._widgets.items():
            if widget == v:
                return k

    def widget_states(
        self, widget: QtWidgets.QWidget | None = None
    ) -> list[Union[DockWidgetState, SplitterState]]:
        states = []

        if widget is None:
            children = self.children()
        else:
            children = [widget.widget(i) for i in range(widget.count())]

        for child in children:
            if isinstance(child, Splitter):
                state = SplitterState(
                    sizes=child.sizes(),
                    orientation=child.orientation(),
                    states=self.widget_states(child),
                )

            elif isinstance(child, DockWidget):
                # widgets is a tuple [title, cls.__name__]
                widgets = [
                    (child.tabText(i), type(child.widget(i)).__name__)
                    for i in range(child.count())
                ]
                state = DockWidgetState(
                    current_index=child.currentIndex(),
                    widgets=widgets,
                    detachable=child.detachable,
                    auto_delete=child.auto_delete,
                    is_center_widget=(child == self.center_widget),
                )
            else:
                # ignore other widget classes
                continue

            if child.isWindow():
                state.geometry = child.geometry()
                state.flags = int(child.windowFlags())

            states.append(state)
        return states

    # noinspection PyMethodMayBeStatic
    def focus_widget(self, widget: QtWidgets.QWidget) -> None:
        parent = widget.parent()
        while parent is not None:
            if isinstance(parent, QtWidgets.QTabWidget):
                break
            parent = parent.parent()
        index = parent.indexOf(widget)
        parent.setCurrentIndex(index)

        window = widget.window()
        window.setWindowState(
            (window.windowState() & ~QtCore.Qt.WindowMinimized) | QtCore.Qt.WindowActive
        )
        window.raise_()
        window.activateWindow()

    def _add_protected_dock_widget(self, parent: Splitter) -> DockWidget:
        dock_widget = DockWidget(self, self)
        dock_widget.auto_delete = False
        parent.addWidget(dock_widget)
        parent.setCollapsible(parent.count() - 1, False)
        return dock_widget

    def _add_widget(
        self, widget: QtWidgets.QWidget | type | str, title: str | None = None
    ) -> tuple[str, QtWidgets.QWidget]:
        # adds a new widget to the window
        if isinstance(widget, str):
            key = widget
            registered_widget = self.registered_widgets.get(key)
            if not registered_widget:
                raise ValueError(f'widget for {repr(key)} has not been registered')
            widget = registered_widget.cls()
        else:
            if isinstance(widget, type):
                cls = widget
                widget = cls()
            else:
                cls = type(widget)
            registered_widget = self.register_widget(cls, title)
            key = cls.__name__

        # check if widget is unique
        if registered_widget.unique:
            if any(type(widget) is type(w) for w in self._widgets.values()):
                raise ValueError(
                    f'Widget {repr(key)} can only be added once to DockWindow '
                )
        # get title
        if title is None:
            title = registered_widget.title

        titles = list(self._widgets.keys())
        unique_title = helper.unique_name(title, titles)

        widget.destroyed.connect(lambda: self._widget_destroy(widget))
        self._widgets[unique_title] = widget
        self.widget_added.emit(widget)

        return unique_title, widget

    def _dock_rects(
        self,
    ) -> OrderedDict[DockWidget, dict[QtCore.Qt.DockWidgetArea, QtCore.QRect]]:
        rects = OrderedDict()
        widgets = self.dock_widgets()
        for widget in widgets:
            rects[widget] = widget.dock_rects()
        return rects

    def _set_widget_states(
        self,
        states: list[DockWidgetState | SplitterState],
        parent: QtWidgets.QWidget,
        widgets: dict[str, QtWidgets.QWidget],
    ) -> None:
        # inner loop for setting widget states
        # removes widgets from widgets dict that have been added to the new layout

        for i, state in enumerate(states):
            # create widget
            if isinstance(state, SplitterState):
                if parent == self:
                    splitter = self.center_splitter
                    splitter.setOrientation(state.orientation)
                else:
                    splitter = Splitter(state.orientation)
                self._set_widget_states(state.states, splitter, widgets)
                splitter.setSizes(state.sizes)

                widget = splitter
            elif isinstance(state, DockWidgetState):
                if state.is_center_widget:
                    dock_widget = self.center_widget
                else:
                    dock_widget = DockWidget(self)
                    dock_widget.detachable = state.detachable
                    dock_widget.auto_delete = state.auto_delete

                for title, cls_name in state.widgets:
                    # remove widget from dictionary to keep track of
                    # which widgets have been re-parented
                    widget = widgets.pop(title, None)

                    if widget is None:
                        try:
                            title, widget = self._add_widget(cls_name)
                        except ValueError:
                            continue
                    dock_widget.addTab(widget, title)

                dock_widget.setCurrentIndex(state.current_index)

                widget = dock_widget
            else:
                continue

            # parent widget
            if isinstance(parent, Splitter):
                if parent.widget(i) is not None:
                    # not replacing widget with itself prevents warnings
                    if parent.widget(i) != widget:
                        parent.replaceWidget(i, widget)
                        widget.setParent(parent)
                        widget.show()
                else:
                    parent.addWidget(widget)
            else:
                widget.show()

            # set window
            if state.flags:
                flags = QtCore.Qt.WindowFlags(state.flags)
                widget.setWindowFlags(flags)
                widget.setGeometry(state.geometry)
                widget.show()

    def _widget_destroy(self, widget: QtWidgets.QWidget) -> None:
        # NOTE: At the point destroyed() is emitted, the widget isn't a QWidget anymore,
        #   just a QObject (as destroyed() is emitted from QObject)
        title = self.widget_title(widget)
        if title is not None:
            # self.widget_removed.emit(widget)
            self._widgets.pop(title)


def area_orientation(area: QtCore.Qt.DockWidgetArea) -> QtCore.Qt.Orientation:
    if area in (QtCore.Qt.LeftDockWidgetArea, QtCore.Qt.RightDockWidgetArea):
        return QtCore.Qt.Horizontal
    elif area in (QtCore.Qt.TopDockWidgetArea, QtCore.Qt.BottomDockWidgetArea):
        return QtCore.Qt.Vertical



from __future__ import annotations

import copy
import dataclasses
from typing import Any, Callable

from PySide2 import QtGui, QtCore, QtWidgets

from qt_extensions.helper import title
from qt_extensions.icons import MaterialIcon


def set_flag(item: QtGui.QStandardItem, flag: QtCore.Qt.ItemFlag, value: bool) -> None:
    if value:
        item.setFlags(item.flags() | flag)
    else:
        item.setFlags(item.flags() & ~flag)


def check_flag(
    item: QtGui.QStandardItem | QtCore.QModelIndex, flag: QtCore.Qt.ItemFlag
) -> bool:
    return bool(item.flags() & flag)


@dataclasses.dataclass
class Field:
    name: str = ''
    label: str | None = None
    editable: bool = False

    def __post_init__(self) -> None:
        if self.label is None:
            self.label = title(self.name)


class ElementModel(QtGui.QStandardItemModel):
    element_added: QtCore.Signal = QtCore.Signal(object)
    element_changed: QtCore.Signal = QtCore.Signal(object, object)
    element_moved: QtCore.Signal = QtCore.Signal(object, QtCore.QModelIndex)
    element_removed: QtCore.Signal = QtCore.Signal(object)

    def __init__(
        self, fields: list[Field] | None = None, parent: QtWidgets.QWidget | None = None
    ) -> None:
        super().__init__(parent)

        # this is used to track elements about to be moved
        self.selected_indexes = []

        if not fields:
            fields = [Field('name')]
        self.fields = fields
        self.refresh_header()

    def clear(self):
        super().clear()
        self.refresh_header()

    def dropMimeData(
        self,
        data: QtCore.QMimeData,
        action: QtCore.Qt.DropAction,
        row: int,
        column: int,
        parent: QtCore.QModelIndex,
    ) -> bool:
        result = super().dropMimeData(data, action, row, 0, parent.siblingAtColumn(0))
        if result and action == QtCore.Qt.MoveAction:
            for index in self.selected_indexes:
                element = index.data(QtCore.Qt.UserRole)
                self.element_moved.emit(element, parent)
            self.selected_indexes = []
        return result

    def removeRow(
        self, row: int, parent: QtCore.QModelIndex = QtCore.QModelIndex()
    ) -> bool:
        index = self.index(row, 0, parent)
        if index.isValid():
            self.element_removed.emit(index.data(QtCore.Qt.UserRole))
        return super().removeRow(row, parent)

    def setData(
        self,
        index: QtCore.QModelIndex,
        value: Any,
        role: int = QtCore.Qt.EditRole,
    ) -> Any:
        result = super().setData(index, value, role)
        if result and role & QtCore.Qt.DisplayRole | QtCore.Qt.EditRole:
            element_index = index.siblingAtColumn(0)
            if element_index.isValid():
                element = index.data(QtCore.Qt.UserRole)
                previous = copy.deepcopy(element)

                try:
                    field = self.fields[index.column()]
                except KeyError:
                    return False

                element = self._set_value(element, value, field)
                self.element_changed.emit(element, previous)
        return result

    def append_element(
        self,
        element: Any = None,
        icon: QtGui.QIcon | None = None,
        movable: bool = True,
        no_children: bool = False,
        parent: QtCore.QModelIndex | None = None,
    ) -> QtCore.QModelIndex:
        # get parent QStandardItem
        parent_item = self.itemFromIndex(parent) if parent else None
        if parent_item is None:
            parent_item = self.invisibleRootItem()

        # create QStandardItems
        items = []
        for field in self.fields:
            value = self._value(element, field)
            item = QtGui.QStandardItem(value)
            item.setEditable(field.editable and movable)
            item.setDragEnabled(movable)
            item.setDropEnabled(not no_children)
            set_flag(item, QtCore.Qt.ItemNeverHasChildren, no_children)
            items.append(item)

        # update QStandardItems
        if items:
            item = items[0]
            item.setData(element, QtCore.Qt.UserRole)
            if icon:
                item.setIcon(icon)
            parent_item.appendRow(items)
            self.element_added.emit(element)
            return item.index()
        return QtCore.QModelIndex()

    def duplicate_index(self, index: QtCore.QModelIndex) -> QtCore.QModelIndex:
        element = index.data(QtCore.Qt.UserRole)
        copied_element = copy.deepcopy(element)
        icon = index.data(QtCore.Qt.DecorationRole)
        movable = check_flag(index, QtCore.Qt.ItemIsDragEnabled)
        no_children = check_flag(index, QtCore.Qt.ItemNeverHasChildren)
        parent = index.parent()
        copied_index = self.append_element(
            copied_element, icon, movable, no_children, parent
        )
        return copied_index

    def element(self, index: QtCore.QModelIndex) -> Any:
        data = self.data(index.siblingAtColumn(0), QtCore.Qt.UserRole)
        return data

    def elements(self, parent: QtCore.QModelIndex | None = None) -> list:
        if parent is None:
            parent = QtCore.QModelIndex()
        elements = []
        for row in range(self.rowCount(parent)):
            index = self.index(row, 0, parent)
            if not index.isValid():  # optimization
                continue
            data = index.data(QtCore.Qt.UserRole)
            if data is not None:
                elements.append(data)
            elements.extend(self.elements(index))
        return elements

    def find_indexes(
        self,
        value: Any,
        field: Field | None = None,
        parent: QtCore.QModelIndex | None = None,
    ) -> list[QtCore.QModelIndex]:
        indexes = []

        if parent is None or value is None:
            parent = QtCore.QModelIndex()

        for row in range(self.rowCount(parent)):
            index = self.index(row, 0, parent)
            if not index.isValid():  # optimization
                continue

            data = index.data(QtCore.Qt.UserRole)
            if field is None and value == data:
                indexes.append(index)
            elif value == self._value(data, field):
                indexes.append(index)

            indexes.extend(self.find_indexes(value, field, index))
        return indexes

    def refresh_index(self, index: QtCore.QModelIndex) -> None:
        element = self.element(index)
        for column, field in enumerate(self.fields):
            item_index = index.siblingAtColumn(column)
            value = self._value(element, field)
            self.setData(item_index, value, QtCore.Qt.DisplayRole)
        # refresh child indexes
        # for row in range(self.rowCount(index)):
        #     self.refresh_index(self.index(row, 0, index))

    def refresh_element(self, element: Any) -> None:
        # update the DisplayRole based on the element stored in the first item
        for index in self.find_indexes(element):
            self.refresh_index(index)

    def refresh_header(self) -> None:
        labels = [field.label for field in self.fields]
        self.setHorizontalHeaderLabels(labels)

    def _value(self, element: Any, field: Field) -> Any:
        if isinstance(element, (str, int, float)):
            value = element
        elif isinstance(element, dict):
            value = element.get(field.name)
        elif isinstance(element, (list, tuple)):
            try:
                i = self.fields.index(field)
                value = element[i]
            except (KeyError, ValueError):
                value = None
        else:
            try:
                value = getattr(element, field.name)
            except AttributeError:
                value = None
        return value

    def _set_value(self, element: Any, value: Any, field: Field) -> Any:
        # the reason for returning element is in case element is an immutable object
        if isinstance(element, (str, int, float)):
            element = value
        elif isinstance(element, dict):
            element[field.name] = value
        elif isinstance(element, (list, tuple)):
            try:
                i = self.fields.index(field)
                element[i] = value
            except (KeyError, ValueError):
                pass
        else:
            try:
                setattr(element, field.name, value)
            except AttributeError:
                pass
        return element


class ElementProxyModel(QtCore.QSortFilterProxyModel):
    # autoAcceptChildRows is a Qt6 feature
    _autoAcceptChildRows = False

    def autoAcceptChildRows(self) -> bool:
        return self._autoAcceptChildRows

    def setAutoAcceptChildRows(self, value: bool):
        self._autoAcceptChildRows = value

    def filterAcceptsRow(
        self, source_row: int, source_parent: QtCore.QModelIndex
    ) -> bool:
        if super().filterAcceptsRow(source_row, source_parent):
            return True
        if self.autoAcceptChildRows() and source_parent.isValid():
            source_row = source_parent.row()
            source_parent = source_parent.parent()
            return self.filterAcceptsRow(source_row, source_parent)
        return False

    def lessThan(
        self, source_left: QtCore.QModelIndex, source_right: QtCore.QModelIndex
    ) -> bool:
        # sort elements and groups separately
        is_left_group = check_flag(source_left, QtCore.Qt.ItemNeverHasChildren)
        is_right_group = check_flag(source_right, QtCore.Qt.ItemNeverHasChildren)

        if is_left_group == is_right_group:
            result = super().lessThan(source_left, source_right)
            return result
        else:
            return is_right_group


class ElementDelegate(QtWidgets.QStyledItemDelegate):
    def paint(
        self,
        painter: QtGui.QPainter,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ) -> None:
        if not check_flag(index, QtCore.Qt.ItemIsDragEnabled):
            option.font.setBold(True)
        option.state &= ~QtWidgets.QStyle.State_HasFocus
        super().paint(painter, option, index)

    def sizeHint(
        self, option: QtWidgets.QStyleOptionViewItem, index: QtCore.QModelIndex
    ) -> QtCore.QSize:
        size_hint = super().sizeHint(option, index)
        size_hint = size_hint.grownBy(QtCore.QMargins(0, 4, 0, 4))
        return size_hint


class ElementTree(QtWidgets.QTreeView):
    selection_changed: QtCore.Signal = QtCore.Signal()

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self.class_groups = True
        self.selected_indexes = []

        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.setSelectionBehavior(QtWidgets.QTableView.SelectRows)
        self.setSortingEnabled(True)
        self.setAutoExpandDelay(500)
        self.setItemDelegate(ElementDelegate())

    def dragMoveEvent(self, event: QtGui.QDragMoveEvent) -> None:
        if self.class_groups:
            # don't allow drag onto items with different element classes
            index = self.indexAt(event.pos())
            parent_element = index.data(QtCore.Qt.UserRole)
            if parent_element is not None:
                for index in self.selected_indexes:
                    element = index.data(QtCore.Qt.UserRole)
                    if not isinstance(element, type(parent_element)):
                        event.ignore()
                        return

        super().dragMoveEvent(event)

    def selectionChanged(
        self, selected: QtCore.QItemSelection, deselected: QtCore.QItemSelection
    ) -> None:
        # store selected indexes
        indexes = []
        model = self.model()
        if model:
            indexes = self.selectionModel().selectedRows()
            if model.sourceModel():
                indexes = [model.mapToSource(index) for index in indexes]
                model = model.sourceModel()

            # used for tracking elements about to be moved
            model.selected_indexes = indexes

        self.selected_indexes = indexes

        self.selection_changed.emit()
        super().selectionChanged(selected, deselected)

    def selected_elements(self) -> list:
        elements = [index.data(QtCore.Qt.UserRole) for index in self.selected_indexes]
        return elements

    def resize_columns(self) -> None:
        if self.model():
            self.expandAll()
            for column in range(self.model().columnCount()):
                self.resizeColumnToContents(column)
            self.collapseAll()


class ElementBrowser(QtWidgets.QWidget):
    selection_changed: QtCore.Signal = QtCore.Signal()

    def __init__(
        self, fields: list[Field] | None = None, parent: QtWidgets.QWidget | None = None
    ) -> None:
        super().__init__(parent)

        self._fields = fields
        self._actions = {}

        self._init_model()
        self._init_ui()

    def _init_model(self):
        self.model = ElementModel(self._fields, self)
        self.proxy = ElementProxyModel(self)
        self.proxy.setAutoAcceptChildRows(True)
        self.proxy.setDynamicSortFilter(False)
        self.proxy.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)
        self.proxy.setFilterKeyColumn(-1)
        self.proxy.setRecursiveFilteringEnabled(True)
        self.proxy.setSortCaseSensitivity(QtCore.Qt.CaseInsensitive)
        self.proxy.setSourceModel(self.model)

    def _init_ui(self):
        self.setLayout(QtWidgets.QVBoxLayout())

        self.toolbar = QtWidgets.QToolBar()
        icon_size = self.toolbar.style().pixelMetric(QtWidgets.QStyle.PM_SmallIconSize)
        self.toolbar.setIconSize(QtCore.QSize(icon_size, icon_size))
        self.layout().addWidget(self.toolbar)

        self.add_toolbar_action(
            'add_element',
            icon=MaterialIcon('add'),
            slot=self.add_element,
        )
        self.add_toolbar_action(
            'remove',
            icon=MaterialIcon('remove'),
            slot=self.remove_selected,
        )
        self.add_toolbar_action(
            'duplicate_element',
            icon=MaterialIcon('content_copy'),
            slot=self.duplicate_selected,
        )
        self.add_toolbar_action(
            'add_group',
            icon=MaterialIcon('create_new_folder'),
            slot=self.add_group,
        )

        filter_bar = QtWidgets.QLineEdit()
        filter_bar.setPlaceholderText('Filter...')
        filter_bar.textChanged.connect(self.filter)
        action = self.toolbar.addWidget(filter_bar)
        action.setText('Filter')

        self.tree = ElementTree()
        self.tree.setModel(self.proxy)
        self.tree.selection_changed.connect(self.selection_changed.emit)
        self.tree.selection_changed.connect(self._update_action_states)
        if not self._fields:
            self.tree.setHeaderHidden(True)
        self.layout().addWidget(self.tree)

        self._update_action_states()

    def filter(self, text: str) -> None:
        self.tree.collapseAll()
        self.proxy.setFilterWildcard(text)
        if text:
            self.tree.expandAll()

    def add_element(self) -> QtCore.QModelIndex:
        element = 'Unnamed'
        parent = self._current_parent()
        index = self.model.append_element(element, no_children=True, parent=parent)
        return index

    def add_group(self) -> QtCore.QModelIndex:
        element = 'Unnamed'
        parent = self._current_parent()
        index = self.model.append_element(
            element, icon=MaterialIcon('folder'), parent=parent
        )
        return index

    def add_toolbar_action(
        self,
        name: str,
        label: str | None = None,
        icon: QtGui.QIcon | None = None,
        slot: Callable | None = None,
    ) -> QtWidgets.QAction:
        if name in self._actions:
            raise ValueError(f'Action with name {name} already exists.')
        if label is None:
            label = title(name)

        action = QtWidgets.QAction(self)
        if label:
            action.setText(label)
        if icon:
            action.setIcon(icon)
        if slot:
            action.triggered.connect(slot)
        self.toolbar.addAction(action)
        self._actions[name] = action
        return action

    def duplicate_selected(self) -> list[QtCore.QModelIndex]:
        indexes = []
        for index in self.tree.selected_indexes:
            copied_index = self.model.duplicate_index(index)
            indexes.append(copied_index)
        return indexes

    def elements(self) -> list:
        return self.model.elements()

    def remove_toolbar_action(self, name: str) -> None:
        if name in self._actions:
            action = self._actions.pop(name)
            self.toolbar.removeAction(action)

    def remove_selected(self) -> None:
        # build list of persistent indexes so that indexes don't changed
        # as rows are removed
        persistent_indexes = [
            QtCore.QPersistentModelIndex(index) for index in self.tree.selected_indexes
        ]
        for index in persistent_indexes:
            if index.isValid() and check_flag(index, QtCore.Qt.ItemIsDragEnabled):
                self.model.removeRow(index.row(), index.parent())

    def select_elements(self, elements: list) -> None:
        selection_model = self.tree.selectionModel()
        selection_model.clear()
        for element in elements:
            indexes = self.model.find_indexes(element)
            for index in indexes:
                index = self.proxy.mapFromSource(index)
                selection_model.select(index, QtCore.QItemSelectionModel.Select)

    def selected_elements(self) -> list:
        return self.tree.selected_elements()

    def _current_parent(self) -> QtCore.QModelIndex:
        indexes = self.tree.selected_indexes
        if indexes:
            index = indexes[0]
            if check_flag(index, QtCore.Qt.ItemNeverHasChildren):
                index = index.parent()
        else:
            index = QtCore.QModelIndex()
        return index

    def _update_action_states(self) -> None:
        indexes = self.tree.selected_indexes
        no_children = False
        movable = True
        if indexes:
            no_children = check_flag(indexes[0], QtCore.Qt.ItemNeverHasChildren)
            movable = check_flag(indexes[0], QtCore.Qt.ItemIsDragEnabled)

        try:
            self._actions['remove'].setEnabled(bool(indexes) and movable)
        except KeyError:
            pass

        try:
            self._actions['duplicate_element'].setEnabled(bool(indexes) and no_children)
        except KeyError:
            pass



from __future__ import annotations

from PySide2 import QtWidgets, QtGui

from qt_extensions.icons import MaterialIcon


class MessageBox(QtWidgets.QMessageBox):
    @staticmethod
    def critical(
        parent: QtWidgets.QWidget | None,
        title: str,
        text: str,
        buttons: QtWidgets.QMessageBox.StandardButton = QtWidgets.QMessageBox.Ok,
        default_button: QtWidgets.QMessageBox.StandardButton = QtWidgets.QMessageBox.NoButton,
    ) -> QtWidgets.QMessageBox.StandardButton:
        icon = MaterialIcon('error')
        color = QtGui.QColor('#ff1744')
        return MessageBox.message(
            parent, title, text, buttons, default_button, icon, color
        )

    @staticmethod
    def information(
        parent: QtWidgets.QWidget | None,
        title: str,
        text: str,
        buttons: QtWidgets.QMessageBox.StandardButton = QtWidgets.QMessageBox.Ok,
        default_button: QtWidgets.QMessageBox.StandardButton = QtWidgets.QMessageBox.NoButton,
    ) -> QtWidgets.QMessageBox.StandardButton:
        icon = MaterialIcon('info')
        return MessageBox.message(parent, title, text, buttons, default_button, icon)

    @staticmethod
    def question(
        parent: QtWidgets.QWidget | None,
        title: str,
        text: str,
        buttons: QtWidgets.QMessageBox.StandardButton = QtWidgets.QMessageBox.Yes
        | QtWidgets.QMessageBox.No,
        default_button: QtWidgets.QMessageBox.StandardButton = QtWidgets.QMessageBox.NoButton,
    ) -> QtWidgets.QMessageBox.StandardButton:
        icon = MaterialIcon('help')
        return MessageBox.message(parent, title, text, buttons, default_button, icon)

    @staticmethod
    def warning(
        parent: QtWidgets.QWidget | None,
        title: str,
        text: str,
        buttons: QtWidgets.QMessageBox.StandardButton = QtWidgets.QMessageBox.Ok,
        default_button: QtWidgets.QMessageBox.StandardButton = QtWidgets.QMessageBox.NoButton,
    ) -> QtWidgets.QMessageBox.StandardButton:
        icon = MaterialIcon('warning')
        return MessageBox.message(parent, title, text, buttons, default_button, icon)

    @staticmethod
    def message(
        parent: QtWidgets.QWidget | None,
        title: str,
        text: str,
        buttons: QtWidgets.QMessageBox.StandardButton = QtWidgets.QMessageBox.Ok,
        default_button: QtWidgets.QMessageBox.StandardButton = QtWidgets.QMessageBox.NoButton,
        icon: MaterialIcon | None = None,
        color: QtGui.QColor | None = None,
    ) -> QtWidgets.QMessageBox.StandardButton:
        message_box = QtWidgets.QMessageBox(parent)
        message_box.setWindowTitle(title)
        message_box.setText(text)
        message_box.setStandardButtons(buttons)
        message_box.setDefaultButton(default_button)
        if icon:
            if parent is None:
                parent = QtWidgets.QApplication.instance()
            style = parent.style()
            size = style.pixelMetric(QtWidgets.QStyle.PM_MessageBoxIconSize)
            pixmap = icon.pixmap(size, color=color)
            message_box.setIconPixmap(pixmap)
        return message_box.exec_()
